//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Vector/RotMatrix.h
//! @brief     Declares class RotMatrix.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_BASE_VECTOR_ROTMATRIX_H
#define BORNAGAIN_BASE_VECTOR_ROTMATRIX_H

#include <heinz/Vectors3D.h>
#include <array>

//! Rotation matrix in three dimensions.

class RotMatrix {
public:
    enum ERotationType { EULER, XAXIS, YAXIS, ZAXIS };

    //! Constructs unit transformation
    RotMatrix();

    //! Destructor
    ~RotMatrix() = default;

    //! Clones the transformation
    RotMatrix* clone() const;

    //! Creates rotation around x-axis
    static RotMatrix createRotateX(double phi);

    //! Creates rotation around y-axis
    static RotMatrix createRotateY(double phi);

    //! Creates rotation around z-axis
    static RotMatrix createRotateZ(double phi);

    //! Creates rotation defined by Euler angles
    static RotMatrix createRotateEuler(double alpha, double beta, double gamma);

    //! Calculates the Euler angles corresponding to the rotation
    std::array<double,3> zxzEulerAngles() const;

    //! Calculates the rotation angle for a rotation around the x-axis alone
    //! Only meaningfull if the actual rotation is around the x-axis
    double calculateRotateXAngle() const;

    //! Calculates the rotation angle for a rotation around the y-axis alone
    //! Only meaningfull if the actual rotation is around the y-axis
    double calculateRotateYAngle() const;

    //! Calculates the rotation angle for a rotation around the z-axis alone
    //! Only meaningfull if the actual rotation is around the z-axis
    double calculateRotateZAngle() const;

    //! Returns the inverse transformation.
    RotMatrix getInverse() const;

    //! Return transformed vector _v_.
    template <class T>
    T transformed(const T& v) const;

    //! Return transformed vector _v_.
    template <class T>
    T transformedInverse(const T& v) const;

    //! Composes two transformations
    RotMatrix operator*(const RotMatrix&) const;

    //! Provides equality operator
    bool operator==(const RotMatrix&) const;

    //! Retrieve the rotation type (general, around x, y or z-axis)
    ERotationType getRotationType() const;

    //! Determine if the transformation is trivial (identity)
    bool isIdentity() const;

    bool isXRotation() const;
    bool isYRotation() const;
    bool isZRotation() const;

private:
    RotMatrix(double x_, double y_, double z_, double s_);

    double x, y, z, s;
};

#endif // BORNAGAIN_BASE_VECTOR_ROTMATRIX_H
