//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/IOFactory.h
//! @brief     Defines class IOFactory.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_HISTO_IOFACTORY_H
#define BORNAGAIN_DEVICE_HISTO_IOFACTORY_H

#include <functional>
#include <string>

template <class T>
class Powerfield;
class IHistogram;
class SimulationResult;

//! Provides users with possibility to read and write IntensityData from/to files
//! in different format. Type of the file will be deduced from file name.
//! *.txt - ASCII file with 2D array [nrow][ncol], layout as in numpy.
//! *.int - BornAgain internal ASCII format.
//! *.tif - 32-bits tiff file.
//! If file name ends with "*.gz" or "*.bz2" the file will be zipped on the fly using
//! appropriate algorithm.

//! @ingroup input_output

/*! Usage:
\code{.py}
# reading from ASCII file or g-zipped ASCII file
histogram = IOFactory.readIntensityData("filename.txt")
histogram = IOFactory.readIntensityData("filename.txt.gz")

# writing to 32-bits tiff file or b-zipped tiff file
IOFactory.writeIntensityData(histogram, "filename.tif")
IOFactory.writeIntensityData(histogram, "filename.tif.bz2")
\endcode
*/

class IOFactory {
public:
    enum LoaderSelector { automatic, bornagain, tiff, nicos };

    //! Reads file and returns newly created Powerfield object. If selector is automatic, then the
    //! file extension will be used to determine which file type to load. If selector is not
    //! automatic, then this selector will define which type of file to load (no matter which file
    //! extension or content).
    //! May throw, but will never return nullptr.
    static Powerfield<double>* readPowerfield(const std::string& file_name,
                                              LoaderSelector selector = automatic);

    static Powerfield<double>* readReflectometryData(const std::string& file_name);

    //! Reads file and returns newly created Histogram object
    static IHistogram* readIntensityData(const std::string& file_name);

    //! Writes Powerfield in file
    static void writePowerfield(const Powerfield<double>& data, const std::string& file_name);

    //! Writes histogram in file
    static void writeIntensityData(const IHistogram& histogram, const std::string& file_name);

    //! Writes Powerfield contained in the given SimulationResult object
    static void writeSimulationResult(const SimulationResult& result, const std::string& file_name);

private:
    static Powerfield<double>*
    readPowerfield(const std::string& file_name,
                   std::function<Powerfield<double>*(std::istream&)> readData);

    static void writePowerfield(const std::string& file_name,
                                std::function<void(std::ostream&)> writeData);

    static bool fileTypeMatchesLoaderSelector(const std::string& fileName, LoaderSelector selector);
};

#endif // BORNAGAIN_DEVICE_HISTO_IOFACTORY_H
