//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/SimulationResult.cpp
//! @brief     Implements class SimulationResult.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/SimulationResult.h"
#include "Device/Histo/Histogram2D.h"

SimulationResult::SimulationResult(const Powerfield<double>& data, const ICoordSystem& coords)
    : m_data(data.clone())
    , m_coordsys(coords.clone())
{
    checkRank();
}

SimulationResult::SimulationResult(const Powerfield<double>& data, const ICoordSystem*&& coords)
    : m_data(data.clone())
    , m_coordsys(coords)
{
    checkRank();
}

SimulationResult::~SimulationResult() = default;

SimulationResult::SimulationResult(const SimulationResult& other)
{
    m_data.reset(other.m_data->clone());
    m_coordsys.reset(other.m_coordsys->clone());
}

SimulationResult::SimulationResult(SimulationResult&& other)
    : m_data(std::move(other.m_data))
    , m_coordsys(std::move(other.m_coordsys))
{
}

SimulationResult& SimulationResult::operator=(const SimulationResult& other)
{
    if (&other == this)
        return *this;
    m_data.reset(other.m_data->clone());
    m_coordsys.reset(other.m_coordsys->clone());
    return *this;
}

SimulationResult& SimulationResult::operator=(SimulationResult&& other)
{
    m_data = std::move(other.m_data);
    m_coordsys = std::move(other.m_coordsys);
    return *this;
}

std::unique_ptr<Powerfield<double>> SimulationResult::data(Coords units) const
{
    return std::unique_ptr<Powerfield<double>>(m_coordsys->createConvertedData(*m_data, units));
}

Histogram2D* SimulationResult::histogram2d(Coords units) const
{
    if (m_data->rank() != 2)
        throw std::runtime_error("Error in SimulationResult::histogram2d: "
                                 "dimension of data is not 2. Please use axis(), array() and "
                                 "data() functions for 1D data.");
    auto P_data = data(units);
    return new Histogram2D(*P_data);
}

std::vector<AxisInfo> SimulationResult::axisInfo(Coords units) const
{
    std::vector<AxisInfo> result;
    auto dim = m_coordsys->rank();
    for (size_t i = 0; i < dim; ++i) {
        AxisInfo info = {m_coordsys->axisName(i, units), m_coordsys->calculateMin(i, units),
                         m_coordsys->calculateMax(i, units)};
        result.push_back(info);
    }
    return result;
}

const ICoordSystem& SimulationResult::converter() const
{
    return *m_coordsys;
}

double& SimulationResult::operator[](size_t i)
{
    return (*m_data)[i];
}

const double& SimulationResult::operator[](size_t i) const
{
    return (*m_data)[i];
}

size_t SimulationResult::size() const
{
    return m_data ? m_data->allocatedSize() : 0;
}

SimulationResult SimulationResult::relativeToMaximum() const
{
    std::unique_ptr<Powerfield<double>> data2{m_data->clone()};
    return {data2->normalizedToMaximum(), m_coordsys->clone()};
}

#ifdef BORNAGAIN_PYTHON
PyObject* SimulationResult::array(Coords units) const
{
    std::unique_ptr<Powerfield<double>> data(m_coordsys->createConvertedData(*m_data, units));
    return data->getArray();
}
#endif

std::vector<double> SimulationResult::convertedBinCenters(Coords units) const
{
    return convertedBinCenters(0, units);
}

std::vector<double> SimulationResult::convertedBinCenters(size_t i_axis, Coords units) const
{
    if (i_axis >= m_coordsys->rank())
        throw std::runtime_error(
            "Error in SimulationResult::axis: no axis corresponds to passed index.");
    auto axis = m_coordsys->createConvertedAxis(i_axis, units);
    return axis->binCenters();
}

void SimulationResult::checkRank() const
{
    ASSERT(m_data);
    ASSERT(m_coordsys);
    if (m_data->rank() != m_coordsys->rank())
        throw std::runtime_error("Error in SimulationResults::checkRank(): "
                                 "dimensions of data and unit converter don't match");
}
