//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Correlations/Profiles2D.h
//! @brief     Defines interface class IProfile2D, and children thereof.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES2D_H
#define BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES2D_H

#include "Base/Math/Constants.h"
#include "Base/Types/ICloneable.h"
#include "Param/Node/INode.h"
#ifndef SWIG
#include "Sample/Correlations/IDistribution2DSampler.h"
#endif // SWIG

#ifndef USER_API

//! Interface for two-dimensional distributions in Fourier space.

class IProfile2D : public ICloneable, public INode {
public:
    IProfile2D(const std::vector<double>& PValues);

    IProfile2D* clone() const override = 0;

    double omegaX() const { return m_omega_x; }
    double omegaY() const { return m_omega_y; }
    double decayLengthX() const { return m_omega_x; }
    double decayLengthY() const { return m_omega_y; }
    double gamma() const { return m_gamma; }

    //! Angle in direct space between X- and Y-axis of distribution.
    double delta() const { return M_PI_2; }

    //! Fourier transformed distribution for q in X,Y coordinates
    //! the original distribution (in real space) is assumed to be normalized:
    //! total integral is equal to 1
    virtual double standardizedFT2D(double qx, double qy) const = 0;
    virtual double decayFT2D(double qx, double qy) const = 0;
#ifndef SWIG
    virtual std::unique_ptr<IDistribution2DSampler> createSampler() const = 0;

    //! Creates the Python constructor of this class (or derived classes)
    virtual std::string pythonConstructor() const;
#endif

protected:
    double sumsq(double qx, double qy) const;

    const double& m_omega_x;
    const double& m_omega_y;
    const double& m_gamma;
};

#endif // USER_API

//! Two-dimensional Cauchy distribution in Fourier space;
//! corresponds to a normalized exp(-r) in real space,
//! with \f$r=\sqrt{(\frac{x}{\omega_x})^2 + (\frac{y}{\omega_y})^2}\f$.
//! @ingroup distributionFT

class Profile2DCauchy : public IProfile2D {
public:
    Profile2DCauchy(std::vector<double> P);
    Profile2DCauchy(double omega_x, double omega_y, double gamma);

    Profile2DCauchy* clone() const override;
    std::string className() const final { return "Profile2DCauchy"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"OmegaX", "nm", "Half-width along x axis", 0, INF, 1.},
                {"OmegaY", "nm", "Half-width along y axis", 0, INF, 1.},
                {"Gamma", "rad",
                 "direct-space orientation with respect to the first lattice vector", -M_PI_2,
                 +M_PI_2, 0}};
    }
    double standardizedFT2D(double qx, double qy) const override;
    double decayFT2D(double qx, double qy) const override;
#ifndef SWIG
    std::unique_ptr<IDistribution2DSampler> createSampler() const override;
#endif
};

//! Two-dimensional Gauss distribution in Fourier space;
//! corresponds to normalized exp(-r^2/2) in real space
//! with \f$r=\sqrt{(\frac{x}{\omega_x})^2 + (\frac{y}{\omega_y})^2}\f$.
//! @ingroup distributionFT

class Profile2DGauss : public IProfile2D {
public:
    Profile2DGauss(std::vector<double> P);
    Profile2DGauss(double omega_x, double omega_y, double gamma);

    Profile2DGauss* clone() const override;
    std::string className() const final { return "Profile2DGauss"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"OmegaX", "nm", "Half-width along x axis", 0, INF, 1.},
                {"OmegaY", "nm", "Half-width along y axis", 0, INF, 1.},
                {"Gamma", "rad",
                 "direct-space orientation with respect to the first lattice vector", -M_PI_2,
                 +M_PI_2, 0}};
    }
    double standardizedFT2D(double qx, double qy) const override;
    double decayFT2D(double qx, double qy) const override;
#ifndef SWIG
    std::unique_ptr<IDistribution2DSampler> createSampler() const override;
#endif
};

//! Two-dimensional gate distribution in Fourier space;
//! corresponds to normalized constant if r<1 (and 0 otherwise) in real space,
//! with \f$r=\sqrt{(\frac{x}{\omega_x})^2 + (\frac{y}{\omega_y})^2}\f$.
//! @ingroup distributionFT

class Profile2DGate : public IProfile2D {
public:
    Profile2DGate(std::vector<double> P);
    Profile2DGate(double omega_x, double omega_y, double gamma);

    Profile2DGate* clone() const override;
    std::string className() const final { return "Profile2DGate"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"OmegaX", "nm", "Half-width along x axis", 0, INF, 1.},
                {"OmegaY", "nm", "Half-width along y axis", 0, INF, 1.},
                {"Gamma", "rad",
                 "direct-space orientation with respect to the first lattice vector", -M_PI_2,
                 +M_PI_2, 0}};
    }
    double standardizedFT2D(double qx, double qy) const override;
    double decayFT2D(double qx, double qy) const override;
#ifndef SWIG
    std::unique_ptr<IDistribution2DSampler> createSampler() const override;
#endif
};

//! Two-dimensional cone distribution in Fourier space;
//! corresponds to 1-r if r<1 (and 0 otherwise) in real space
//! with \f$r=\sqrt{(\frac{x}{\omega_x})^2 + (\frac{y}{\omega_y})^2}\f$.
//! @ingroup distributionFT

class Profile2DCone : public IProfile2D {
public:
    Profile2DCone(std::vector<double> P);
    Profile2DCone(double omega_x, double omega_y, double gamma);

    Profile2DCone* clone() const override;
    std::string className() const final { return "Profile2DCone"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"OmegaX", "nm", "Half-width along x axis", 0, INF, 1.},
                {"OmegaY", "nm", "Half-width along y axis", 0, INF, 1.},
                {"Gamma", "rad",
                 "direct-space orientation with respect to the first lattice vector", -M_PI_2,
                 +M_PI_2, 0}};
    }
    double standardizedFT2D(double qx, double qy) const override;
    double decayFT2D(double qx, double qy) const override;
#ifndef SWIG
    std::unique_ptr<IDistribution2DSampler> createSampler() const override;
#endif
};

//! Two-dimensional Voigt distribution in Fourier space;
//! corresponds to eta*Gauss + (1-eta)*Cauchy
//! @ingroup distributionFT

class Profile2DVoigt : public IProfile2D {
public:
    Profile2DVoigt(std::vector<double> P);
    Profile2DVoigt(double omega_x, double omega_y, double gamma, double eta);

    Profile2DVoigt* clone() const override;
    std::string className() const final { return "Profile2DVoigt"; }
    // const auto tooltip = "Pseudo-Voigt distribution";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"OmegaX", "nm", "Half-width along x axis", 0, INF, 1.},
                {"OmegaY", "nm", "Half-width along y axis", 0, INF, 1.},
                {"Gamma", "rad",
                 "direct-space orientation with respect to the first lattice vector", -M_PI_2,
                 +M_PI_2, 0},
                {"Eta", "", "balances between Gauss (eta=0) and Lorentz (eta=1) limiting cases", 0,
                 1, .5}};
    }
    double standardizedFT2D(double qx, double qy) const override;
    double decayFT2D(double qx, double qy) const override;
    double eta() const { return m_eta; }
#ifndef SWIG
    std::unique_ptr<IDistribution2DSampler> createSampler() const override;
    std::string pythonConstructor() const override;
#endif

protected:
    const double& m_eta;
};

#endif // BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES2D_H
