//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ISimulation2D.h
//! @brief     Defines interface ISimulation2D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_ISIMULATION2D_H
#define BORNAGAIN_SIM_SIMULATION_ISIMULATION2D_H

#include "Sim/Simulation/ISimulation.h"
#include <memory>

class Beam;
class IDetector;
class DetectorContext;
class DiffuseElement;
class IDetector;
class IShape2D;

//! Abstract base class of simulations that generate 2D patterns.
//!
//! Base class of OffspecSimulation and ScatteringSimulation.
//!
//! Holds the common implementations for simulations with a 2D detector.

class ISimulation2D : public ISimulation {
public:
    ISimulation2D(const Beam& beam, const MultiLayer& sample, const IDetector& detector);
    ISimulation2D(const MultiLayer& sample);
    ~ISimulation2D() override;

    std::vector<const INode*> nodeChildren() const override;

    //! Adds mask of given shape to the stack of detector masks. The mask value 'true' means
    //! that the channel will be excluded from the simulation. The mask which is added last
    //! has priority.
    //! @param shape The shape of mask (Rectangle, Polygon, Line, Ellipse)
    //! @param mask_value The value of mask
    void addMask(const IShape2D& shape, bool mask_value = true);

    //! Put the mask for all detector channels (i.e. exclude whole detector from the analysis)
    void maskAll();

    //! Sets rectangular region of interest with lower left and upper right corners defined.
    void setRegionOfInterest(double xlow, double ylow, double xup, double yup);

    bool force_polarized() const override;

    Beam& beam() { return *m_beam; }
    IDetector& detector() { return *m_detector; }

#ifndef SWIG

    const Beam& beam() const { return *m_beam; }
    const IDetector& detector() const { return *m_detector; }
    const IDetector* getDetector() const { return m_detector.get(); }

protected:
    //! Put into a clean state for running a simulation
    void prepareSimulation() override;

    virtual void initCoordSystem() {}

    void initDistributionHandler() override;

    //! Gets the number of elements this simulation needs to calculate
    size_t numberOfElements() const override;

    //! Generate a single threaded computation for a given range of simulation elements
    //! @param re_sample Preprocessed version of our sample
    //! @param start Index of the first element to include into computation
    //! @param n_elements Number of elements to process
    std::unique_ptr<IComputation> createComputation(const reSample& re_sample, size_t start,
                                                    size_t n_elements) override;

    //! Generate simulation elements for given beam
    std::vector<std::unique_ptr<DiffuseElement>> generateElements(const Beam& beam);

    //! Normalize the detector counts to beam intensity, to solid angle, and to exposure angle.
    //! @param start_ind Index of the first element to operate on
    //! @param n_elements Number of elements to process
    void normalize(size_t start_ind, size_t n_elements) override;

    void addBackgroundIntensity(size_t start_ind, size_t n_elements) override;

    void addDataToCache(double weight) override;

    void moveDataFromCache() override;

    std::unique_ptr<Beam> m_beam;
    std::unique_ptr<IDetector> m_detector;
    std::vector<std::unique_ptr<DiffuseElement>> m_eles;
    std::vector<double> m_cache;

private:
    std::unique_ptr<DetectorContext> m_detector_context;
#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_ISIMULATION2D_H
