#include "Device/Data/LLData.h"
#include "Tests/GTestWrapper/google_test.h"
#include <algorithm>
#include <heinz/Complex.h>

class LLDataTest : public ::testing::Test {
protected:
    LLDataTest();
    ~LLDataTest() override;

    LLData<int>* int_data_0d;
    LLData<float>* fl_data_1d;
    LLData<double>* db_data_3d;
};

LLDataTest::LLDataTest()
{
    auto* dim0 = new int[0];

    auto* dim1 = new int[1];
    dim1[0] = 10;

    auto* dim3 = new int[3];
    dim3[0] = 10;
    dim3[1] = 15;
    dim3[2] = 20;

    int_data_0d = new LLData<int>(0u, dim0);
    fl_data_1d = new LLData<float>(1u, dim1);
    db_data_3d = new LLData<double>(3u, dim3);

    delete[] dim0;
    delete[] dim1;
    delete[] dim3;
}

LLDataTest::~LLDataTest()
{
    delete int_data_0d;
    delete fl_data_1d;
    delete db_data_3d;
}


TEST_F(LLDataTest, TotalSize)
{
    EXPECT_EQ(1u, int_data_0d->getTotalSize());
    EXPECT_EQ(10u, fl_data_1d->getTotalSize());
    EXPECT_EQ(3000u, db_data_3d->getTotalSize());
}

TEST_F(LLDataTest, GetRank)
{
    EXPECT_EQ(0u, int_data_0d->rank());
    EXPECT_EQ(1u, fl_data_1d->rank());
    EXPECT_EQ(3u, db_data_3d->rank());
}

TEST_F(LLDataTest, SetAll)
{
    db_data_3d->setAll(1.0);
    EXPECT_DOUBLE_EQ((*db_data_3d)[0], 1.0);
}

TEST_F(LLDataTest, TotalSum)
{
    fl_data_1d->setAll(2.0);
    EXPECT_FLOAT_EQ(fl_data_1d->getTotalSum(), 20.0);

    db_data_3d->setAll(0.5);
    EXPECT_DOUBLE_EQ(db_data_3d->getTotalSum(), 1500.0);
}

TEST_F(LLDataTest, GetDimensions)
{
    EXPECT_EQ(int_data_0d->sizes(), (int*)nullptr);
    EXPECT_EQ(fl_data_1d->sizes()[0], 10);
    EXPECT_EQ(db_data_3d->sizes()[1], 15);
}

TEST_F(LLDataTest, DataCopyingConstructor)
{
    auto* other_int_data_0d = new LLData<int>(*int_data_0d);
    EXPECT_TRUE(HaveSameSizes(*int_data_0d, *other_int_data_0d));

    fl_data_1d->setAll(1.2);
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    EXPECT_TRUE(HaveSameSizes(*fl_data_1d, *other_fl_data_1d));
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 1.2f);
    EXPECT_FLOAT_EQ(fl_data_1d->getTotalSum(), other_fl_data_1d->getTotalSum());

    db_data_3d->setAll(1.17);
    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    EXPECT_TRUE(HaveSameSizes(*db_data_3d, *other_db_data_3d));
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[10], 1.17);
    EXPECT_DOUBLE_EQ(db_data_3d->getTotalSum(), other_db_data_3d->getTotalSum());

    delete other_int_data_0d;
    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, DataAssignment)
{
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    fl_data_1d->setAll(1.1f);
    (*other_fl_data_1d) = (*fl_data_1d);
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 1.1f);
    EXPECT_FLOAT_EQ(fl_data_1d->getTotalSum(), other_fl_data_1d->getTotalSum());

    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    db_data_3d->setAll(1.27);
    (*other_db_data_3d) = (*db_data_3d);
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[11], 1.27);
    EXPECT_DOUBLE_EQ(db_data_3d->getTotalSum(), other_db_data_3d->getTotalSum());

    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, Addition)
{
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    fl_data_1d->setAll(1.1f);
    other_fl_data_1d->setAll(2.9f);
    (*other_fl_data_1d) += (*fl_data_1d);
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 4.0f);
    EXPECT_FLOAT_EQ(other_fl_data_1d->getTotalSum(), 40.0f);

    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    db_data_3d->setAll(1.27);
    other_db_data_3d->setAll(0.73);
    (*other_db_data_3d) += (*db_data_3d);
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[11], 2.0);
    EXPECT_DOUBLE_EQ(other_db_data_3d->getTotalSum(), 6000.0);

    auto* odim1 = new int[1];
    odim1[0] = 12;
    auto* fl_data_1d2 = new LLData<float>(1u, odim1);
    fl_data_1d2->setAll(10.0);

    EXPECT_THROW((*fl_data_1d) += (*fl_data_1d2), std::runtime_error);

    delete fl_data_1d2;
    delete[] odim1;

    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, Substraction)
{
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    fl_data_1d->setAll(1.15f);
    other_fl_data_1d->setAll(5.15f);
    (*other_fl_data_1d) -= (*fl_data_1d);
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 4.0f);
    EXPECT_FLOAT_EQ(other_fl_data_1d->getTotalSum(), 40.0f);

    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    db_data_3d->setAll(1.27);
    other_db_data_3d->setAll(1.77);
    (*other_db_data_3d) -= (*db_data_3d);
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[11], 0.5);
    EXPECT_DOUBLE_EQ(other_db_data_3d->getTotalSum(), 1500.0);

    auto* odim1 = new int[1];
    odim1[0] = 12;
    auto* fl_data_1d2 = new LLData<float>(1u, odim1);
    fl_data_1d2->setAll(10.0);

    EXPECT_THROW((*fl_data_1d) -= (*fl_data_1d2), std::runtime_error);

    delete fl_data_1d2;
    delete[] odim1;

    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, Multiplication)
{
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    fl_data_1d->setAll(1.15f);
    other_fl_data_1d->setAll(2.0f);
    (*other_fl_data_1d) *= (*fl_data_1d);
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 2.30f);
    EXPECT_FLOAT_EQ(other_fl_data_1d->getTotalSum(), 23.0f);

    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    db_data_3d->setAll(1.25);
    other_db_data_3d->setAll(2.0);
    (*other_db_data_3d) *= (*db_data_3d);
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[11], 2.5);
    EXPECT_DOUBLE_EQ(other_db_data_3d->getTotalSum(), 7500.0);

    auto* odim1 = new int[1];
    odim1[0] = 12;
    auto* fl_data_1d2 = new LLData<float>(1u, odim1);
    fl_data_1d2->setAll(10.0);

    EXPECT_THROW((*fl_data_1d) *= (*fl_data_1d2), std::runtime_error);

    delete fl_data_1d2;
    delete[] odim1;

    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, Division)
{
    auto* other_fl_data_1d = new LLData<float>(*fl_data_1d);
    fl_data_1d->setAll(1.15f);
    other_fl_data_1d->setAll(2.3f);
    (*other_fl_data_1d) /= (*fl_data_1d);
    EXPECT_FLOAT_EQ((*other_fl_data_1d)[0], 2.0f);
    EXPECT_FLOAT_EQ(other_fl_data_1d->getTotalSum(), 20.0f);

    auto* other_db_data_3d = new LLData<double>(*db_data_3d);
    db_data_3d->setAll(1.25);
    other_db_data_3d->setAll(2.5);
    (*other_db_data_3d) /= (*db_data_3d);
    EXPECT_DOUBLE_EQ((*other_db_data_3d)[11], 2.0);
    EXPECT_DOUBLE_EQ(other_db_data_3d->getTotalSum(), 6000.0);

    auto* odim1 = new int[1];
    odim1[0] = 12;
    auto* fl_data_1d2 = new LLData<float>(1u, odim1);
    fl_data_1d2->setAll(10.0);

    EXPECT_THROW((*fl_data_1d) /= (*fl_data_1d2), std::runtime_error);

    delete fl_data_1d2;
    delete[] odim1;
    delete other_fl_data_1d;
    delete other_db_data_3d;
}

TEST_F(LLDataTest, HaveSameSizes)
{
    auto* odim0 = new int[0];

    auto* odim1 = new int[1];
    odim1[0] = 10;

    auto* odim3 = new int[3];
    odim3[0] = 10;
    odim3[1] = 15;
    odim3[2] = 20;

    auto* other_int_data_0d = new LLData<int>(0u, odim0);
    auto* other_fl_data_1d = new LLData<float>(1u, odim1);
    auto* other_db_data_3d = new LLData<double>(3u, odim3);

    EXPECT_TRUE(HaveSameSizes(*int_data_0d, *other_int_data_0d));
    EXPECT_TRUE(HaveSameSizes(*fl_data_1d, *other_fl_data_1d));
    EXPECT_TRUE(HaveSameSizes(*db_data_3d, *other_db_data_3d));

    odim3[1] = 25;
    auto* some_other_db_data_3d = new LLData<double>(3u, odim3);
    EXPECT_FALSE(HaveSameSizes(*db_data_3d, *some_other_db_data_3d));

    delete other_int_data_0d;
    delete other_fl_data_1d;
    delete other_db_data_3d;
    delete some_other_db_data_3d;
    delete[] odim0;
    delete[] odim1;
    delete[] odim3;
}

TEST_F(LLDataTest, Accessors)
{
    for (size_t i = 0; i < fl_data_1d->getTotalSize(); ++i)
        (*fl_data_1d)[i] = 0.5f * i;
    EXPECT_FLOAT_EQ((*fl_data_1d)[2], 1.0f);

    EXPECT_FLOAT_EQ(fl_data_1d->getTotalSum(), 22.5f);

    for (size_t i = 0; i < db_data_3d->getTotalSize(); ++i)
        (*db_data_3d)[i] = 0.2 * i;

    EXPECT_DOUBLE_EQ((*db_data_3d)[2], 0.4);

    EXPECT_DOUBLE_EQ(db_data_3d->getTotalSum(), 899700.0);
}

TEST_F(LLDataTest, Max)
{
    db_data_3d->setAll(15.);
    (*db_data_3d)[13] = 60;
    EXPECT_EQ(db_data_3d->max(), 60);
}
